module huber_reg_mod
  
  implicit none
  
  real,dimension(:) , pointer :: dath
  real              , private :: gamdh
  real              , private :: gamxh
  real              , private :: epsi
  integer           , private :: nreg
  integer           , private :: l1
  integer           , private :: verb
  logical, parameter, private :: T = .true., F = .false.
  
  
contains
  
  subroutine huber_init_out(gamd_in,gamx_in,l1_in,eps_in,nreg_in,verb_in)
    
    real                    :: gamd_in,gamx_in,eps_in
    integer                 :: verb_in, nreg_in, l1_in
    
    gamdh  = gamd_in
    gamxh  = gamx_in
    epsi   = eps_in
    verb   = verb_in
    nreg   = nreg_in
    l1     = l1_in

    if (verb>1) then
       if (epsi.ne.0.) then
          if (l1.eq.0) write(0,*) '- Regularization with the Huber norm'
          if (l1.eq.1) write(0,*) '- Regularization with the Cauchy norm'
          if (l1.eq.2) write(0,*) '- Regularization with the l2 norm'
       end if
    end if

  end subroutine huber_init_out

  subroutine huber_init_in(datah)
    
    real,dimension(:)       :: datah
    
    allocate(dath(size(datah)))
    
    dath  = datah
    
  end subroutine huber_init_in
  
  function huberfct(oper,reg,x,fct) result (stat)

    interface

       function oper( adj, add, x, dat) result (stat)
         integer             :: stat
         logical, intent(in) :: adj, add
         real, dimension(: ) :: x, dat
       end function oper
       
       function reg( adj, add, x, dat) result (stat)
         integer             :: stat
         logical, intent(in) :: adj, add
         real, dimension(: ) :: x, dat
       end function reg
    
    end interface

    integer                            :: stat,i
    real, dimension( :),intent( in)    :: x             
    real,               intent(out)    :: fct      
    real, dimension(size(dath))        :: rd
    real, dimension(nreg)              :: rm
    real                               :: tmp

    rd = - dath

    stat  = oper(F,T,x,rd)               ! r_d = L x - data

    where(abs(rd).gt.gamdh)
       rd = abs(rd)-gamdh/2.0
    elsewhere
       rd = rd*rd/(2.0*gamdh)
    endwhere

    fct = sum(rd)

    if (epsi.ne.0.0) then
       
       tmp=0.
       stat = reg(F,F,x,rm) 
       
       if (l1.eq.0) then

          ! Huber
          where(abs(rm).gt.gamxh)
             rm = abs(rm)-gamxh/2.0
          elsewhere
             rm = rm*rm/(2.0*gamxh)
          endwhere
       
          tmp = sum(rm)
          
       else if (l1.eq.1) then
          
          do i=1,size(rm)
             tmp = tmp + log(1+rm(i)**2/gamxh**2)
          end do

       else

          rm = rm*rm
          
       end if
       
       fct = epsi*tmp + fct

    end if

  end function huberfct
  
  function hubergdt(oper,reg,x,g) result(stat)
    
    interface
       function oper( adj, add, x, dat) result (stat)
         integer                          :: stat
         logical, intent(in)              :: adj, add
         real, dimension(: )              :: x, dat
       end function oper

       function reg( adj, add, x, dat) result (stat)
         integer                          :: stat
         logical, intent(in)              :: adj, add
         real, dimension(: )              :: x, dat
       end function reg

    end interface
    
    real, dimension( :),intent(in)      :: x
    real, dimension( :),intent(out)     :: g
    real, dimension(size(x))            :: gm
    real, dimension(nreg)               :: rm,zm
    real, dimension(size(dath))         :: rd,zd
    integer                             :: stat, stat1, i
    real                                :: tmp
        
    rd = - dath
    stat = oper(F,T,x,rd)               ! r_d = L x - dat
    
    do i=1,size(rd)   
       tmp = min(1.0,rd(i)/gamdh)
       zd(i) = max(-1.0,tmp)
    end do
    
    stat1 = oper(T,F,g,zd)
    
    if (epsi.ne.0.0) then
        
       stat = reg(F,F,x,rm)

       ! Works for Huber, not sure about Cauchy...
       if (l1.eq.0) then
          
          do i=1,nreg
             tmp = min(1.0,rm(i)/gamxh)  
             zm(i) = max(-1.0,tmp)
          end do
       
       else if (l1.eq.1) then
          
          do i=1,nreg
             zm(i) = (rm(i)/gamxh**2)/(1 + rm(i)**2/gamxh**2)
          end do

       else

          do i=1,nreg
             zm(i)=rm(i)
          end do
       
       end if
       
       stat = reg(T,F,gm,zm)
       
       g = g + epsi*gm
       
    end if
    
  end function hubergdt

  subroutine huber_close()
    deallocate(dath)
  end subroutine huber_close

end module huber_reg_mod

