module parabolic  
  ! parabolic Radon transform in the Fourier domain
  use adj_mod

  implicit none
  integer, private :: nt 
  integer, private :: nx 
  integer, private :: nq 
  integer, private :: alias
  real, private :: t0 
  real, private :: dt 
  real, private :: q0 
  real, private :: dq 
  real, private :: x0 
  real, private :: dx 
  real, private :: fhi
  real, private :: flo
  integer, private :: nthreads

contains
    
  subroutine parabolic_init ( t0_in,dt_in,x0_in,dx_in,q0_in,dq_in,nt_in&
  &,nx_in,nq_in,alias_in,fhi_in,flo_in)

    real    :: t0_in 
    real    :: dt_in 
    real    :: x0_in 
    real    :: dx_in 
    real    :: q0_in 
    real    :: dq_in 
    real    :: fhi_in
    real    :: flo_in

    integer    :: nt_in 
    integer    :: nx_in 
    integer    :: nq_in 
    integer    :: alias_in
    
    t0 = t0_in 
    dt = dt_in 
    x0 = x0_in 
    dx = dx_in 
    q0 = q0_in 
    dq = dq_in 
    nt = nt_in 
    nx = nx_in 
    nq = nq_in
    alias = alias_in
    fhi=fhi_in
    flo=flo_in
    
  end subroutine parabolic_init
  
  function parabolic_lop ( adj, add, modl, data) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: modl,data 
    call adjnull (adj,add,modl,data )
    call parabolic_lop2(adj,modl,data )
    stat=0
  end function parabolic_lop
  
  subroutine parabolic_lop2(adj,modl,data)
    logical,intent(in) :: adj 
    real, dimension (nt,nq)  :: modl 
    real, dimension (nt,nx)  :: data
    complex, allocatable, dimension(:,:) :: cdata
    complex, allocatable, dimension(:,:) :: cmodl
    integer iz,ix,iq,it,iw,stat
    integer :: nw,nwmax,nwmin
    real x,q,w,sx,t,z,wt,w0,dw,pi,fmax
    real taper,wtap,wmin
    complex :: cc
    
    pi=3.14159265

    nw = jpad2(nt) 
    w0 = -pi/dt   ;  dw= 2.*pi/(nw*dt)

    allocate(cdata(nw,nx),cmodl(nw,nq))

    cmodl=cmplx(0.,0.)
    cdata=cmplx(0.,0.)

    if (adj) then
       
       do ix=1,nx       
          cdata(1:nt,ix)=cmplx(data(:,ix),0.)
       end do

       call ft1axis(0,1.,nw,nx,cdata)

    else
       
       do iq=1,nq       
          cmodl(1:nt,iq)=cmplx(modl(:,iq),0.)
       end do

       call ft1axis(0,1.,nw,nq,cmodl)
     
    end if

    do iq= 1, nq

       q = q0 + (iq-1)*dq
       do ix =1 ,nx 
          x = x0 + (ix-1) * dx

          ! antialiasing parameters
          
          if (alias.ne.0) then             

             wmin = pi/(2*abs(dx*x*q))
             wtap = 7*wmin/8
             wmin = 9*wmin/8
           
          end if

          nwmax = fhi*nw*dt+nw/2
          nwmin = flo*nw*dt+nw/2

          do iw= nwmin, nwmax
            
             w = w0 + (iw-1) * dw

             if (alias.ne.0) then

                if (w.le.wtap) then
                   taper=1.
                else if (w.le.wmin) then
                   taper=cos(pi*(w-wtap)/(2*(wmin - wtap)))
                else
                   taper=0.
                end if
                
             else
                
                taper=1.

             end if
                
             if (adj) then
                 
               cc = cexp(cmplx(0.,-w*x*x*q))
               cmodl(iw,iq)= cmodl(iw,iq) + 2*cdata(iw,ix)*cc*taper
                
             else
               
               cc = cexp(cmplx(0.,w*x*x*q))
               cdata(iw,ix)= cdata(iw,ix) + 2*cmodl(iw,iq)*cc*taper
                
             end if
          end do
          
       end do
       
    end do
    
    if (adj) then

       call ft1axis(1,1.,nw,nq,cmodl)

       do iq=1,nq       
          modl(:,iq)=real(cmodl(1:nt,iq))
       end do

    else

       call ft1axis(1,1.,nw,nx,cdata)

       do ix=1,nx       
          data(:,ix)=real(cdata(1:nt,ix))
       end do

    end if

    deallocate(cdata,cmodl)

  end subroutine parabolic_lop2

  integer function jpad2(n)
    integer  n
    jpad2=1
    do
       if (jpad2.ge.n) exit
       jpad2=jpad2*2
    end do
  end function jpad2
    
  subroutine parabolic_close()
  end subroutine parabolic_close
  
end module parabolic
























































































































