module parainv_over 

  ! parabolic Radon transform in the Fourier domain
  ! with pseudo-inverse
  ! Compute (L'L+epsI)-1L'd

  use adj_mod
  use ctoeplitz

  implicit none
  integer, private :: nt 
  integer, private :: nx 
  integer, private :: nq 
  integer, private :: harold
  integer, private :: verb 
  real, private :: t0 
  real, private :: dt 
  real, private :: q0 
  real, private :: dq 
  real, private :: x0 
  real, private :: dx   
  real, private :: eps

contains
    
  subroutine parainvover_init ( t0_in,dt_in,x0_in,dx_in,q0_in,dq_in,nt_in&
  &,nx_in,nq_in,eps_in, verb_in )
      
    real    :: t0_in 
    real    :: dt_in 
    real    :: x0_in 
    real    :: dx_in 
    real    :: q0_in 
    real    :: dq_in 
    real    :: eps_in

    integer    :: nt_in 
    integer    :: nx_in 
    integer    :: nq_in 
    integer    :: verb_in 
    
    t0 = t0_in 
    dt = dt_in 
    x0 = x0_in 
    dx = dx_in 
    q0 = q0_in 
    dq = dq_in 
    nt = nt_in 
    nx = nx_in 
    nq = nq_in
    verb = verb_in

  end subroutine parainvover_init
  
  function parainvover_lop ( adj, add, modl, data) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: modl,data 
    call adjnull (adj,add,modl,data )
    call parainvover_lop2(adj,modl,data )
    stat=0
  end function parainvover_lop
  
  subroutine parainvover_lop2(adj,modl,data)
    logical,intent(in) :: adj 
    real, dimension (nt,nq)  :: modl 
    real, dimension (nt,nx)  :: data
    complex*16, allocatable, dimension(:,:) :: cdata,cmodl
    complex*16, allocatable, dimension(:)   :: toep,rhs,res
    integer    :: iz,ix,iq,it,iw,stat,iq2,nw
    real       :: x,q,w,sx,t,z,wt,w0,dw,pi,q2
    complex*16 :: cc
    
    pi=3.14159265
    
    nw = jpad2(nt) 
    w0 = -pi/dt   ;  dw= 2.*pi/(nw*dt)  
    
    allocate(cdata(nw,nx),cmodl(nw,nq))
    allocate(toep(nq),rhs(nq),res(nq))
    cmodl=cmplx(0.,0.)
    cdata=cmplx(0.,0.)
    
    ! Fourier transform the data

    if (adj) then
       
       do ix=1,nx       
          cdata(1:nt,ix)=cmplx(data(:,ix),0.)
       end do

       call dft1axis(0,1.,nw,nx,cdata)

    else
       
       do iq=1,nq       
          cmodl(1:nt,iq)=cmplx(modl(:,iq),0.)
       end do

       call dft1axis(0,1.,nw,nq,cmodl)

     
    end if

     call ctoeplitz_init(nq)
    
    ! End Fourier transform

    do iw= nw/2,nw
       w = w0 + (iw-1)*dw

       if (verb>0) write(0,*) 'PRT>> frequency=',w/(2.*pi),'/',(w0+(nw-1)*dw)/(2*pi)

       ! Form matrix toep and right hand side (rhs)
       
       toep=cmplx(0.,0.)
        rhs=cmplx(0.,0.)
        res=cmplx(0.,0.)
       
       do iq =1 ,nq
          q = q0 + (iq-1) * dq
          
          do ix= 1, nx  
             x = x0 + (ix-1) * dx
             
             if (adj) then
                
                ! just the first row is needed since toep is hermitian
                toep(iq)=toep(iq)+cexp(cmplx(0.,w*x*x*dq*(iq-1)))

                ! compute the right hand side L'd
                cc = cexp(cmplx(0.,-w*x*x*q))
                rhs(iq)=rhs(iq)+cdata(iw,ix)*cc
             
             else
                
                cc = cexp(cmplx(0.,w*x*x*q))
                cdata(iw,ix)= cdata(iw,ix) + 2*cmodl(iw,iq)*cc
                
             end if
             
          end do
          
       end do
       
       if (adj) then
          
          toep(1)=toep(1)+eps  ! toep = L'L+epsI
          call ctoeplitz_solve(toep,rhs,res)
          cmodl(iw,:)=2*res

       end if
       
    end do
        
    call ctoeplitz_close()

    if (adj) then
       
       call dft1axis(1,1.,nw,nq,cmodl)
       
       do iq=1,nq
          modl(:,iq)=real(cmodl(1:nt,iq))
       end do
           
    else
       
       call dft1axis(1,1.,nw,nx,cdata)
       
       do ix=1,nx 
          data(:,ix)=real(cdata(1:nt,ix))
       end do
       
    end if

    deallocate(cdata,cmodl)
    deallocate(toep,rhs,res)

  end subroutine parainvover_lop2
  
  integer function jpad2(n)
    integer n
    jpad2=1
    do
       if (jpad2.ge.n) exit
       jpad2=jpad2*2
    end do
  end function jpad2
  
end module parainv_over









