module mult_nmo_util

use sep
use morgan_util

implicit none

contains

!--------------------------------------------------------------------------------
! Compute total extra zero-offset traveltime of 1st and 2nd order multiples
!--------------------------------------------------------------------------------
real function get_taumult( y1, y2, x1, x2, xp, ny1, ny2, dy1, dy2, y01,     &
                           y02, tauw, korder, isplit, isurf, nsurf )        &
                                                             result(tauw_tot)

	real,    intent(IN) :: y1, y2, x1, x2, xp, dy1, dy2, y01, y02
	integer, intent(IN) :: korder, isplit, ny1, ny2, isurf, nsurf
	real                :: yr1(2), yr2(2)
	integer             :: iy, jy, iy1, iy2, jy1, jy2

	real, dimension(:), intent(IN) :: tauw

	if( korder==0 ) then                                   ! Primary reflection
		tauw_tot = 0.0
	else if( korder==1 ) then                            ! First-order multiple
		call get_mult1_yrefl( yr1, xp, x1, x2, y1, y2, isplit)
		iy1 = min( ny1, max(1,nint( (yr1(1)-y01)/dy1 + 1 )) )
		iy2 = min( ny2, max(1,nint( (yr1(2)-y02)/dy2 + 1 )) )
		iy = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,isurf/) )
		tauw_tot = tauw(iy)
	else if( korder==2) then                            ! Second-order multiple
		call get_mult2_yrefl( yr1, yr2, xp, x1, x2, y1, y2, isplit)
		iy1 = min( ny1, max(1,nint( (yr1(1)-y01)/dy1 + 1 )) )
		iy2 = min( ny2, max(1,nint( (yr1(2)-y02)/dy1 + 1 )) )
		jy1 = min( ny1, max(1,nint( (yr2(1)-y01)/dy1 + 1 )) )
		jy2 = min( ny2, max(1,nint( (yr2(2)-y02)/dy1 + 1 )) )
		iy = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,isurf/) )
		jy = grid2line3( (/ny1,ny2,nsurf/), (/jy1,jy2,isurf/) )
		tauw_tot = tauw(iy) + tauw(jy)
	else
		call seperr("ERROR: get_tauw(): multiple order > 2 not yet implemented")
	end if

end function get_taumult
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! Compute total (spatially-variant) reflection coefficient of 1st and 2nd 
! order multiples 
!--------------------------------------------------------------------------------
real function get_refl( y1, y2, x1, x2, xp, ny1, ny2, dy1, dy2, y01, y02,   &
                        refl, korder, isplit, isurf, nsurf, invrefl )       &
                                                             result(refl_tot)

	real,    intent(IN) :: y1, y2, x1, x2, xp, dy1, dy2, y01, y02
	integer, intent(IN) :: korder, isplit, ny1, ny2, isurf, nsurf
	logical, intent(IN) :: invrefl
	real                :: yr1(2), yr2(2)
	integer             :: iy, jy, iy1, iy2, jy1, jy2

	real, dimension(:), intent(IN) :: refl

	if( korder==0 ) then                                   ! Primary reflection
		refl_tot = 1.0
	else if( korder==1 ) then                            ! First-order multiple
		call get_mult1_yrefl( yr1, xp, x1, x2, y1, y2, isplit)
		iy1 = min( ny1, max(1,nint( (yr1(1)-y01)/dy1 + 1 )) )
		iy2 = min( ny2, max(1,nint( (yr1(2)-y02)/dy2 + 1 )) )
		iy = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,isurf/) )
		refl_tot = refl(iy)
	else if( korder==2) then                            ! Second-order multiple
		call get_mult2_yrefl( yr1, yr2, xp, x1, x2, y1, y2, isplit)
		iy1 = min( ny1, max(1,nint( (yr1(1)-y01)/dy1 + 1 )) )
		iy2 = min( ny2, max(1,nint( (yr1(2)-y02)/dy2 + 1 )) )
		jy1 = min( ny1, max(1,nint( (yr2(1)-y01)/dy1 + 1 )) )
		jy2 = min( ny2, max(1,nint( (yr2(2)-y02)/dy2 + 1 )) )
		iy = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,isurf/) )
		jy = grid2line3( (/ny1,ny2,nsurf/), (/jy1,jy2,isurf/) )
		refl_tot = refl(iy)*refl(jy)
	else
		call seperr("ERROR: get_refl(): multiple order > 2 not yet implemented")
	end if

	if( invrefl .AND. abs(refl_tot) > 1.0e-6)  refl_tot = 1./refl_tot
	if( invrefl .AND. abs(refl_tot) < 1.0e-6)  refl_tot = 1.

end function get_refl
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! Compute squared effective velocity for first and second order multiples.
!--------------------------------------------------------------------------------
!    Ve^2 = [ tauprim*Vprim^2 + SUM( taumult(i)*Vmult(i)^2 ) ]
!         / [ tauprim + SUM( taumult(i) ) ]
!--------------------------------------------------------------------------------
real function get_veff2( y1, y2, x1, x2, xp, tau, ny1, ny2, dy1, dy2, y01, y02, &
                         nt, dt, t0, vel, tauw, korder, isplit, isurf, nsurf )  &
                                                                    result(veff2)

	real,    intent(IN) :: y1, y2, x1, x2, xp, tau, dy1, dy2, y01, y02, dt, t0
	integer, intent(IN) :: korder, isplit, ny1, ny2, nt, nsurf, isurf
	real                :: ymr1(2), ymr2(2), ypr(2), vmult1, vmult2, vprim,       &
	                       taumult1, taumult2
	integer             :: iy1, iy2, j, itau, iy

	real, dimension(:), intent(IN) :: vel, tauw

	call get_prim_yrefl( y1, y2, xp, x1, x2, ypr, isplit, korder)
	iy1 = min( ny1, max(1,nint( (ypr(1)-y01)/dy1 + 1 )) )
	iy2 = min( ny2, max(1,nint( (ypr(2)-y02)/dy2 + 1 )) )
	itau = min( nt, max(1,nint( (tau-t0)/dt + 1 )) )
	j = grid2line3( (/nt,ny1,ny2/), (/itau,iy1,iy2/) )
	vprim = vel(j)*vel(j)                                     ! vprim

	if( korder==0 ) then                                    ! veff2(prim) = veff2
		veff2 = vprim
	else if( korder==1 ) then                                        ! multiple 1
		call get_mult1_yrefl( ymr1, xp, x1, x2, y1, y2, isplit)
		iy1 = min( ny1, max(1,nint( (ymr1(1)-y01)/dy1 + 1 )) )
		iy2 = min( ny2, max(1,nint( (ymr1(2)-y02)/dy2 + 1 )) )
		iy = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,isurf/) )
		taumult1 = tauw(iy)
		itau = min( nt, max(1,nint( (taumult1-t0)/dt+1 )) )
		j = grid2line3( (/nt,ny1,ny2/), (/itau,iy1,iy2/) )
		vmult1 = vel(j)*vel(j)                   
		veff2 = ( taumult1*vmult1 + tau*vprim )/( taumult1 + tau) 
	else if( korder==2 ) then                                          ! multiple 2
		call get_mult2_yrefl( ymr1, ymr2, xp, x1, x2, y1, y2, isplit)
		iy1 = min( ny1, max(1,nint( (ymr1(1)-y01)/dy1 + 1 )) )
		iy2 = min( ny2, max(1,nint( (ymr1(2)-y02)/dy1 + 1 )) )
		iy = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,isurf/) )
		taumult1 = tauw(iy)                                           ! taumult leg 1
		itau = min( nt, max(1,nint((taumult1-t0)/dt+1)) )
		j = grid2line3( (/nt,ny1,ny2/), (/itau,iy1,iy2/) )
		vmult1 = vel(j)*vel(j)                                          ! vmult leg 1
		iy1 = min( ny1, max(1,nint( (ymr2(1)-y01)/dy1 + 1 )) )
		iy2 = min( ny2, max(1,nint( (ymr2(2)-y02)/dy1 + 1 )) )
		iy = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,isurf/) )
		taumult2 = tauw(iy)                                           ! taumult leg 2
		itau = min( nt, max(1,nint( (taumult2-t0)/dt + 1 )) )
		j = grid2line3( (/nt,ny1,ny2/), (/itau,iy1,iy2/) )
		vmult2 = vel(j)*vel(j)                                          ! vmult leg 2
		veff2 = ( taumult1*vmult1 + taumult2*vmult2 + tau*vprim )                   &
		      / ( taumult1+taumult2+tau) 
!write(0,*) 'second order mult: veff/vel',sqrt(veff2),vprim
	else
		call seperr("ERROR: get_veff2(): multiple order > 2 not yet implemented")
	end if

end function get_veff2
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! Event following in 3-D
! Assume input dips are in physical units, not grid units.
!--------------------------------------------------------------------------------
real function get_tau3( y1, y2, ytarg1, ytarg2, tau, ny1, dy1, y01, ny2, dy2,   &
                        y02, nt, dt, t0, dip1, dip2, korder, isplit, verb       &
                         ) result(tau3)
	real   ,               intent(IN) :: y1, y2, ytarg1, ytarg2, tau, dy1, dy2,   &
	                                     y01, y02, dt, t0
	integer,               intent(IN) :: korder, isplit, ny1, ny2, nt
	real   , dimension(:), intent(IN) :: dip1, dip2
	real                              :: ytmp1, ytmp2, step1, step2, pix1, pix2
	integer                           :: iy01, iy02, incr, it0, j, jmax, k
	logical                           :: working
	logical                           :: verb
	optional                          :: verb

	it0  = min( nt,  max(1,nint((tau-t0)/dt+1)) )
	iy01 = min( ny1, max(1,nint((y1-y01)/dy1+1)) )
	iy02 = min( ny2, max(1,nint((y2-y02)/dy2+1)) )

if(.not. present(verb)) verb=.false.

	if( korder==0 ) then
		tau3 = tau
	else if( korder==1 ) then
		tau3 = tau
		pix1 = (ytarg1-y1)/dy1
		pix2 = (ytarg2-y2)/dy2
		if( abs(pix1)>0.0 .OR. abs(pix2)>0.0 ) then          ! Only move if needed
			step1 = dy1*pix1/max(abs(pix1),abs(pix2))
			step2 = dy2*pix2/max(abs(pix1),abs(pix2))

if(verb) write(0,"(a,2f10.3)") 'pix1,pix2',pix1, pix2
if(verb) write(0,"(a,4f10.3)") 'y1,y2,ytarg1,ytarg2',y1,y2,ytarg1,ytarg2

			ytmp1 = y1
			ytmp2 = y2	
			jmax = min(50,ceiling(max(abs(pix1),abs(pix2))))
			do j=1,jmax
				iy01 = min( ny1, max(1,nint((ytmp1-y01)/dy1+1)) )    ! Extrapolate dip 
				iy02 = min( ny2, max(1,nint((ytmp2-y02)/dy2+1)) )    ! out from edges.

				if( j==jmax) then
					step1 = ytarg1-ytmp1
					step2 = ytarg2-ytmp2
				end if

				ytmp1 = ytmp1 + step1
				ytmp2 = ytmp2 + step2

				k = grid2line3( (/nt,ny1,ny2/), (/it0,iy01,iy02/) )
if(verb) write(0,"(a,6f10.4,f10.6)") 'ytmp1,ytmp2,p1,p2,step1,step2,dtau',ytmp1,    &
                     ytmp2,dip1(k),dip2(k),step1,step2,step1*dip1(k) + step2*dip2(k)

				tau3 = tau3 + step1*dip1(k) + step2*dip2(k)

			end do
		end if
	else if( korder==2) then
		tau3 = tau
!		if(      isplit==0) then                          ! S10102G
!		else if( isplit==1) then                          ! S10201G
!		else if( isplit==2) then                          ! S20101G
!		end if
!		iyp = min( n3, max(1,iyp) )
	end if

	tau3 = max( 0.0, min( (nt-1)*dt+t0, tau3 ) )

if(verb) write(0,"(a,2f)") 'tau/tau3:',tau,tau3

end function get_tau3
! ------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! Get cosine velocity correction derived from "average" dip angle.
!--------------------------------------------------------------------------------
! Note: currently using vbar=v(1) in mult_nmo_core%tabulate_core().  A 
! conservative velocity estimate, which produces smallest dip
!--------------------------------------------------------------------------------
real function get_vcos( y1, y2, xp, tauprim, ny1, dy1, y01, ny2, dy2, y02, nt,  &
                        dt, t0, x1, x2, dip1, dip2, korder, isplit, isurf,      &
                        nsurf, vbar, tauw )                          result(vcos)

	real   ,               intent(IN) :: y1, y2, xp, tauprim, dy1, dy2, y01, y02, &
	                                     dt, t0, x1, x2, vbar
	integer,               intent(IN) :: ny1, ny2, nt, korder, isplit, isurf, nsurf
	real   , dimension(:), intent(IN) :: dip1, dip2, tauw

	real    :: ypr(2), ymr1(2), ymr2(2), mdipang(2), pdipang, wt, taumult(2),     &
	           xx1, xx2
	integer :: iyp(2), iym1(2), iym2(2), itm1, itm2, itp, j, nw_t, nw_y1, nw_y2,  &
	           iy1, iy2, it

	if( korder>0 ) then
		itp = min( nt, max(1,nint((tauprim-t0)/dt+1)) )
		call get_prim_yrefl( y1, y2, xp, x1, x2, ypr, isplit, korder)
		iyp(1) = min( ny1, max(1,nint((ypr(1)-y01)/dy1+1)) )
		iyp(2) = min( ny2, max(1,nint((ypr(2)-y02)/dy2+1)) )
		if( korder==1 ) then
			call get_mult1_yrefl( ymr1, xp, x1, x2, y1, y2, isplit)
			iym1(1) = min( ny1, max(1,nint((ymr1(1)-y01)/dy1+1)) )
			iym1(2) = min( ny2, max(1,nint((ymr1(2)-y02)/dy2+1)) )
			j = grid2line3( (/ny1,ny2,nsurf/), (/iym1(1),iym1(2),isurf/) )
			taumult(1) = tauw(j)
			itm1 = min( nt, max(1,nint((taumult(1)-t0)/dt+1)) )
		else if( korder==2 ) then
			call get_mult2_yrefl( ymr1, ymr2, xp, x1, x2, y1, y2, isplit)
			iym1(1) = min( ny1, max(1,nint((ymr1(1)-y01)/dy1+1)) )
			iym1(2) = min( ny2, max(1,nint((ymr1(2)-y02)/dy2+1)) )
			iym2(1) = min( ny1, max(1,nint((ymr2(1)-y01)/dy1+1)) )
			iym2(2) = min( ny2, max(1,nint((ymr2(2)-y02)/dy2+1)) )
			j = grid2line3( (/ny1,ny2,nsurf/), (/iym1(1),iym1(2),isurf/) )
			taumult(1) = tauw(j)
			itm1 = min( nt, max(1,nint((taumult(1)-t0)/dt+1)) )
			j = grid2line3( (/ny1,ny2,nsurf/), (/iym2(1),iym2(2),isurf/) )
			taumult(2) = tauw(j)
			itm2 = min( nt, max(1,nint((taumult(2)-t0)/dt+1)) )
		else
			call seperr( "get_vcos(): order>2 not implemented yet")
		end if

		nw_t  = 3
		nw_y1 = 1
		nw_y2 = 1

		wt = 0.0
		pdipang = 0.0
		mdipang = 0.0

		xx1 = abs(x1)
		xx2 = abs(x2)

		do iy2=max(1,iyp(2)-nw_y2),min(ny2,iyp(2)+nw_y2)
			do iy1=max(1,iyp(1)-nw_y1),min(ny1,iyp(1)+nw_y1)
				do it=max(1,itp-nw_t),min(nt,itp+nw_t)
					j = grid2line3( (/nt,ny1,ny2/), (/it,iy1,iy2/) )
					pdipang = pdipang + xx1*abs(dip1(j)) + xx2*abs(dip2(j))
					wt = wt + xx1 + xx2
				end do
			end do
		end do
		if(wt>1e-6) pdipang = pdipang/wt
		pdipang = atan( pdipang*vbar )

		wt = 0.0
		do iy2=max(1,iym1(2)-nw_y2),min(ny2,iym1(2)+nw_y2)
			do iy1=max(1,iym1(1)-nw_y1),min(ny1,iym1(1)+nw_y1)
				do it=max(1,itm1-nw_t),min(nt,itm1+nw_t)
					j = grid2line3( (/nt,ny1,ny2/), (/it,iy1,iy2/) )
					mdipang(1) = mdipang(1) + xx1*abs(dip1(j)) + xx2*abs(dip2(j))
					wt = wt + xx1 + xx2
				end do
			end do
		end do
		if(wt>1e-6) mdipang(1) = mdipang(1)/wt
		mdipang(1) = atan( mdipang(1)*vbar )
	
		if( korder==2 ) then
			wt = 0.0
			do iy2=max(1,iym2(2)-nw_y2),min(ny2,iym2(2)+nw_y2)
				do iy1=max(1,iym2(1)-nw_y1),min(ny1,iym2(1)+nw_y1)
					do it=max(1,itm2-nw_t),min(nt,itm2+nw_t)
						j = grid2line3( (/nt,ny1,ny2/), (/it,iy1,iy2/) )
						mdipang(2) = mdipang(2) + xx1*abs(dip1(j)) + xx2*abs(dip2(j))
						wt = wt + xx1 + xx2
					end do
				end do
			end do
			if(wt>1e-6) mdipang(2) = mdipang(2)/wt
			mdipang(2) = atan( mdipang(2)*vbar )
		end if

	end if

	if( korder==0 ) then
		vcos = 1.0
	else if( korder==1) then
		vcos = cos( pdipang ) / cos( pdipang + mdipang(1) )
!	write(0,*) 'tauprim,taumult(1):',tauprim,taumult(1)
!	write(0,*) 'pdip,mdip,vcos:',pdipang,mdipang(1),vcos,itm1
	else if( korder==2) then
		vcos = cos( pdipang ) / cos( pdipang + mdipang(1) + mdipang(2) )
	end if

	
end function get_vcos
! ------------------------------------------------------------------------------

! ------------------------------------------------------------------------------
! Prepare composite weights for n-linear interpolation, n = 1, 2, or 3.
! ------------------------------------------------------------------------------
subroutine bins_weights_to_1d( n, wt_lo, wt_hi, bin_lo, bin_hi, wt1d, bin1d)
	integer, dimension(:), intent(IN)  :: n, bin_lo, bin_hi
	real,    dimension(:), intent(IN)  :: wt_lo, wt_hi
	integer, dimension(:), intent(OUT) :: bin1d
	real,    dimension(:), intent(OUT) :: wt1d
	integer                            :: k

	k = size(n)

	if( size(wt_lo) /= k ) call seperr("bins_weights_to_1d(): size(wt_lo) wrong")
	if( size(wt_hi) /= k ) call seperr("bins_weights_to_1d(): size(wt_hi) wrong")
	if( size(bin_hi) /= k ) call seperr("bins_weights_to_1d(): size(bin_hi) wrong")
	if( size(bin_lo) /= k ) call seperr("bins_weights_to_1d(): size(bin_lo) wrong")
	if( size(wt1d) /= 2**k ) call seperr("bins_weights_to_1d(): size(wt1d) wrong")
	if( size(bin1d) /= 2**k ) call seperr("bins_weights_to_1d(): size(bin1d) wrong")

	if(k==1) then
		bin1d(1) = bin_lo(1)
		bin1d(2) = bin_hi(1)
		wt1d(1) = wt_lo(1)
		wt1d(2) = wt_hi(1)
	else if(k==2) then
		bin1d(1) = grid2line2( n, (/bin_lo(1),bin_lo(2)/) );                 ! 0,0
		bin1d(2) = grid2line2( n, (/bin_lo(1),bin_hi(2)/) );                 ! 0,1
		bin1d(3) = grid2line2( n, (/bin_hi(1),bin_lo(2)/) );                 ! 1,0
		bin1d(4) = grid2line2( n, (/bin_hi(1),bin_hi(2)/) );                 ! 1,1
		wt1d(1) = wt_lo(1)*wt_lo(2)
		wt1d(2) = wt_lo(1)*wt_hi(2)
		wt1d(3) = wt_hi(1)*wt_lo(2)
		wt1d(4) = wt_hi(1)*wt_hi(2)
	else if(k==3) then
		bin1d(1) = grid2line3( n, (/bin_lo(1),bin_lo(2),bin_lo(3)/) );       ! 0,0,0
		bin1d(2) = grid2line3( n, (/bin_lo(1),bin_lo(2),bin_hi(3)/) );       ! 0,0,1
		bin1d(3) = grid2line3( n, (/bin_lo(1),bin_hi(2),bin_lo(3)/) );       ! 0,1,0
		bin1d(4) = grid2line3( n, (/bin_lo(1),bin_hi(2),bin_hi(3)/) );       ! 0,1,1
		bin1d(5) = grid2line3( n, (/bin_hi(1),bin_lo(2),bin_lo(3)/) );       ! 1,0,0
		bin1d(6) = grid2line3( n, (/bin_hi(1),bin_lo(2),bin_hi(3)/) );       ! 1,0,1
		bin1d(7) = grid2line3( n, (/bin_hi(1),bin_hi(2),bin_lo(3)/) );       ! 1,1,0
		bin1d(8) = grid2line3( n, (/bin_hi(1),bin_hi(2),bin_hi(3)/) );       ! 1,1,1
		wt1d(1) = wt_lo(1)*wt_lo(2)*wt_lo(3)
		wt1d(2) = wt_lo(1)*wt_lo(2)*wt_hi(3)
		wt1d(3) = wt_lo(1)*wt_hi(2)*wt_lo(3)
		wt1d(4) = wt_lo(1)*wt_hi(2)*wt_hi(3)
		wt1d(5) = wt_hi(1)*wt_lo(2)*wt_lo(3)
		wt1d(6) = wt_hi(1)*wt_lo(2)*wt_hi(3)
		wt1d(7) = wt_hi(1)*wt_hi(2)*wt_lo(3)
		wt1d(8) = wt_hi(1)*wt_hi(2)*wt_hi(3)
	else
		call seperr("bins_weights_to_1d(): Case k>3 not yet implemented!")
	end if

end subroutine bins_weights_to_1d
! ------------------------------------------------------------------------------


! ---------------------------------------------------------------------------
real function get_xp( x1, tau, v, tauw, korder, vw, x2) result(xp)
	real   , intent(IN) :: x1, tau, v, tauw, vw
	integer, intent(IN) :: korder
	real                :: ttauk, vv, vvw, vveff, x2, xtot2
	optional            :: x2

	if(.NOT. present(x2)) x2=0.0                                          ! 2-D
	xtot2 = x1*x1 + x2*x2 
	if( korder==0 ) then
		xp = sqrt(xtot2)
	else
		ttauk = korder*tauw + tau
		vv = v*v
		vvw = vw*vw
		vveff = (korder*vvw*tauw + vv*tau)/ttauk
		if( ttauk==0.0 .AND. (xtot2==0.0 .OR. vveff==vv) ) then
			xp = 0.0
		else
			xp = (sqrt(xtot2)*tau*vv)/sqrt( (ttauk*vveff)**2 + xtot2*(vveff-vv) )
		end if
	end if
end function get_xp
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
subroutine get_xp12( xp, x1, x2, xp1, xp2)
	real, intent(IN)  :: xp, x1, x2
	real, intent(OUT) :: xp1, xp2
	real              :: theta, eps

	eps = 1.0e-6
	if( abs(x1)<eps .AND. abs(x2)<eps ) then
		xp1=0.0; xp2=0.0
	else if( abs(x1)<eps .AND. abs(x2)>eps ) then
		xp2 = sign(xp,x2);  xp1=0.0	
	else if( abs(x1)>eps .AND. abs(x2)<eps ) then
		xp1 = sign(xp,x1);  xp2=0.0
	else
		theta = atan(x2/x1)
		xp1 = xp*cos(theta)
		xp2 = xp*sin(theta)
	end if

end subroutine get_xp12
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
! Get reflection point (midpoint coords) of target reflection
! ---------------------------------------------------------------------------
subroutine get_prim_yrefl( y1, y2, xp, x1, x2, yr, isplit, korder)
	real,    intent(IN)  :: xp, x1, x2, y1, y2
	integer, intent(IN)  :: isplit, korder
	real,    intent(OUT) :: yr(2)
	real                 :: xp1, xp2

	if( korder==0 ) then                                              ! Primary
		yr(1)=y1;  yr(2)=y2
	else if( korder==1 ) then                            ! First-order multiple
		call get_xp12( xp, x1, x2, xp1, xp2)
		if( isplit==0) then                                               ! S102G
			yr(1) = y1 + (x1-xp1)/2.0;  yr(2) = y2 + (x2-xp2)/2.0
		else if( isplit==1) then                                          ! S201G
			yr(1) = y1 - (x1-xp1)/2.0;  yr(2) = y2 - (x2-xp2)/2.0
		end if
	else if( korder==2 ) then                           ! Second-order multiple
		call get_xp12( xp, x1, x2, xp1, xp2)
		if( isplit==0) then                                             ! S10102G
			yr(1) = y1 + (x1-xp1);  yr(2) = y2 + (x2-xp2)
		else if( isplit==1) then                                        ! S10201G
			yr(1)=y1;  yr(2)=y2
		else if( isplit==2) then                                        ! S20101G
			yr(1) = y1 - (x1-xp1);  yr(2) = y2 - (x2-xp2)
		end if
	else                               ! higher-order multiples not implemented
		call seperr("ERROR: get_prim_yrefl(): mult order > 2 not implemented")
	end if

end subroutine get_prim_yrefl
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
! Get reflection point (midpoint coords) of multiple legs of first-order mult
! ---------------------------------------------------------------------------
subroutine get_mult1_yrefl( yr, xp, x1, x2, y1, y2, isplit)
	integer, intent(IN) :: isplit
	real, intent(IN)    :: xp, x1, x2, y1, y2
	real, intent(OUT)   :: yr(2)
	real                :: xp1, xp2

	call get_xp12( xp, x1, x2, xp1, xp2)

	if( isplit==0) then                                               ! S102G
		yr(1) = y1 - xp1/2.0
		yr(2) = y2 - xp2/2.0
	else if( isplit==1) then                                          ! S201G
		yr(1) = y1 + xp1/2.0
		yr(2) = y2 + xp2/2.0
	else
		call seperr("ERROR: get_mult1_yrefl(): Only isplit=0,1 for mult 1")
	end if

end subroutine get_mult1_yrefl
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
! Get reflection point (midpoint coords) of multiple legs of 2nd-order mult
! ---------------------------------------------------------------------------
subroutine get_mult2_yrefl( yr1, yr2, xp, x1, x2, y1, y2, isplit)
	integer, intent(IN) :: isplit
	real, intent(IN)    :: xp, x1, x2, y1, y2
	real, intent(OUT)   :: yr1(2), yr2(2)
	real                :: xp1, xp2

	call get_xp12( xp, x1, x2, xp1, xp2)

	if( isplit==0) then                                               ! S10102G
		yr1(1) = y1 - xp1/4.0 - x1/4.0
		yr1(2) = y2 - xp2/4.0 - x2/4.0
		yr2(1) = y1 - 3.0*xp1/4.0 + x1/4.0
		yr2(2) = y2 - 3.0*xp2/4.0 + x2/4.0
	else if( isplit==1) then                                          ! S10201G
		yr1(1) = y1 - xp1/4.0 - x1/4.0
		yr1(2) = y2 - xp2/4.0 - x2/4.0
		yr2(1) = y1 + xp1/4.0 + x1/4.0
		yr2(2) = y2 + xp2/4.0 + x2/4.0
	else if( isplit==2) then                                          ! S20101G
		yr1(1) = y1 + 3.0*xp1/4.0 - x1/4.0
		yr1(2) = y2 + 3.0*xp2/4.0 - x2/4.0
		yr2(1) = y1 + xp1/4.0 + x1/4.0
		yr2(2) = y2 + xp2/4.0 + x2/4.0
	else
		call seperr("ERROR: get_mult1_yrefl(): Only isplit=0,1 for mult 1")
	end if

end subroutine get_mult2_yrefl
! ---------------------------------------------------------------------------

!--------------------------------------------------------------------------------
subroutine get_nullcube( nt, dt, t0, nx1, dx1, x01, nx2, dx2, x02, ny1, dy1,    &
                         y01, ny2, dy2, y02, iy1, iy2, maxorder, nmod, nsurf,   &
                         tauw, vel, nullcube                                    )

	real,    intent(IN) :: dt, t0, dx1, x01, dx2, x02, dy1, y01, dy2, y02
	integer, intent(IN) :: nt, nx1, nx2, ny1, ny2, iy1, iy2, maxorder, nmod, nsurf 
	integer             :: iord, isplit, isurf, nnsurf, itauw, jtauw, ix1, ix2,   &
	                       it, k, i0, j0, j1, ixsnell
	real                :: tmax, xmax, x1, x2, xtot2, tau, veff2, tbnd, xp, y1, y2

	real,    dimension(:), intent(IN)  :: tauw, vel
	integer, dimension(:), intent(OUT) :: nullcube


	tmax = (nt-1)*dt + t0
	xmax = (nx1-1)*dx1 + x01
	y1 = (iy1-1)*dy1 + y01
	y2 = (iy2-1)*dy2 + y02

	do iord=0,maxorder
		if(iord==0) then
			nnsurf = 1
		else
			nnsurf = nsurf
		end if
		do isurf=1,nnsurf
			do isplit=0,iord

				itauw = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,isurf/) )
				jtauw = nint( (tauw(itauw)-t0)/dt + 1 )

				do ix1=1,nx1
					x1 = (ix1-1)*dx1+x02
					do ix2=1,nx2
						x2 = (ix2-1)*dx2+x02
						xtot2 = x1*x1 + x2*x2
						do it=1,nt
							k = get_cubeloc( iord, nsurf, isurf, isplit)
							i0 = grid2line4( (/nt,nx1,nx2,nmod/), (/it,ix1,ix2,k/) )

							if( it < jtauw - 10 ) then
								nullcube(i0) = 0
							else
								nullcube(i0) = 1
							end if

							tau = (it-1)*dt+t0
							j0 = grid2line3( (/nt,ny1,ny2/), (/jtauw,iy1,iy2/) )     ! V(tauw)
							j1 = grid2line3( (/nt,ny1,ny2/), (/it,iy1,iy2/)    )     ! V(tau)
							xp = get_xp( xmax, tau, vel(j1), tauw(itauw), iord, vel(j0) )
							veff2 = get_veff2( y1, y2, x1, x2, xp, tau, ny1, ny2, dy1, dy2,   &
							                   y01, y02, nt, dt, t0, vel, tauw, iord, isplit, &
							                   isurf, nsurf                                   )

							tbnd = sqrt( tau**2 + xtot2/veff2 )

							if( tbnd < tmax ) then
								nullcube(i0) = nullcube(i0)*1
							else
								nullcube(i0) = nullcube(i0)*0
							end if

							ixsnell = nint( (xp-x01)/dx1 + 1 ) + 2

							if( ix1 < ixsnell ) then
								nullcube(i0) = nullcube(i0)*1
							else
								nullcube(i0) = nullcube(i0)*0
							end if

						end do
					end do
				end do

			end do
		end do
	end do

end subroutine get_nullcube
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! Automatically-generated residual weight.  Zero above WBR.  Detect missing 
! traces by auto-division.
!--------------------------------------------------------------------------------
subroutine get_reswt( nt, dt, t0, nx1, dx1, x01, nx2, dx2, x02, ny1, ny2, iy1,  &
                      iy2, nsurf, data, tauw, vel, reswt                        )
	real,    intent(IN) :: dt, t0, dx1, x01, dx2, x02
	integer, intent(IN) :: nt, nx1, nx2, ny1, ny2, iy1, iy2, nsurf 
	integer             :: itauw, jtauw, ix1, ix2, it, i0, j1
	real                :: x1, x2, xtot2, v2, tbnd
	real, dimension(:), intent(IN)  :: data, tauw, vel
	real, dimension(:), intent(OUT) :: reswt

	itauw = grid2line3( (/ny1,ny2,nsurf/), (/iy1,iy2,1/) )
	jtauw = nint( (tauw(itauw)-t0)/dt + 1 )

	reswt = 1.0

	do ix1=1,nx1
		x1 = (ix1-1)*dx1+x02
		do ix2=1,nx2
			x2 = (ix2-1)*dx2+x02
			xtot2 = x1*x1 + x2*x2
			j1 = grid2line3( (/nt,ny1,ny2/), (/jtauw,iy1,iy2/) )               ! V(tau)
			v2 = vel(j1)*vel(j1)

			tbnd = sqrt( (tauw(itauw)-10*dt)**2 + xtot2/v2 )

			do it=1,nt
				if( (it-1)*dt+t0 < tbnd) then
					i0 = grid2line3( (/nt,nx1,nx2/), (/it,ix1,ix2/) )
					reswt(i0) = 0.0
				end if
			end do
		end do
	end do

	where( data==0.0 ) reswt = 0.0

end subroutine get_reswt
!--------------------------------------------------------------------------------

! ---------------------------------------------------------------------------
! Linear interpolation weights
! ---------------------------------------------------------------------------
subroutine get_interp1_weights( pt, d, o, bin_lo, bin_hi, wt_lo, wt_hi,     &
                                bnd_lo, bnd_hi                              )
	real   , intent(IN)  :: pt, d, o
	integer, intent(IN)  :: bnd_lo, bnd_hi
	real   , intent(OUT) :: wt_lo, wt_hi
	integer, intent(OUT) :: bin_lo, bin_hi
	real                 :: ipt

	ipt = (pt-o)/d + 1
	bin_lo = floor(ipt);  bin_hi = bin_lo + 1
	wt_lo = bin_hi - ipt; wt_hi = 1.0 - wt_lo

	if( bin_lo <= bnd_lo ) then
		if( bin_lo == bnd_lo) then
			wt_lo = 1.0;  wt_hi = 0.0              ! On edge
		else
			wt_lo = 0.0;  wt_hi = 0.0              ! Out of bounds
		end if
		bin_lo = bnd_lo
		bin_hi = bnd_lo + 1
	end if

	if( bin_hi >= bnd_hi ) then
		if( bin_hi == bnd_hi ) then
			wt_lo = 0.0;  wt_hi = 1.0              ! On edge
		else
			wt_lo = 0.0;  wt_hi = 0.0              ! Out of bounds
		end if
		bin_lo = bnd_hi - 1
		bin_hi = bnd_hi
	end if
end subroutine get_interp1_weights
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
! n-linear interpolation weights
! ---------------------------------------------------------------------------
subroutine get_interpn_weights( pt, d, o, bin_lo, bin_hi, wt_lo, wt_hi,     &
                                bnd_lo, bnd_hi                              )

	real   , dimension(:)              :: pt, d, o
	real   , dimension(:)              :: wt_lo, wt_hi
	integer, dimension(:)              :: bin_lo, bin_hi, bnd_lo, bnd_hi
	real   , dimension(:), allocatable :: ipt
	integer                            :: n, i

	n = size(pt)
	if( n<2 ) call seperr("get_interpn_weights(): use get_interp1_weights()")

	if(      size(d) /= n) call seperr("get_interpn_weights(): size(d) wrong")
	if(      size(o) /= n) call seperr("get_interpn_weights(): size(o) wrong")
	if( size(bin_lo) /= n) call seperr("get_interpn_weights(): size(bin_lo) wrong")
	if( size(bin_hi) /= n) call seperr("get_interpn_weights(): size(bin_hi) wrong")
	if(  size(wt_hi) /= n) call seperr("get_interpn_weights(): size(wt_hi) wrong")
	if(  size(wt_lo) /= n) call seperr("get_interpn_weights(): size(wt_lo) wrong")
	if( size(bnd_lo) /= n) call seperr("get_interpn_weights(): size(bnd_lo) wrong")
	if( size(bnd_hi) /= n) call seperr("get_interpn_weights(): size(bnd_hi) wrong")

	allocate(ipt(n))
	ipt = (pt-o)/d + 1.0
	bin_lo = floor(ipt);  bin_hi = bin_lo + 1
	wt_lo = bin_hi - ipt; wt_hi = 1.0 - wt_lo
	deallocate(ipt)

	do i=1,n
		if( bin_lo(i) < bnd_lo(i) ) then
			if( bnd_hi(i)==1 .AND. bnd_lo(i)==1 ) then         ! Check for n(i) = 1
				wt_lo(i) = 0.5;  bin_lo(i) = 1
				wt_hi(i) = 0.5;  bin_hi(i) = 1
			else
				wt_lo(i) = 0.0;  wt_hi(i) = 0.0
				bin_lo(i) = bnd_lo(i)
				bin_hi(i) = bnd_lo(i) + 1
			end if
		end if

		if( bin_hi(i) > bnd_hi(i) ) then
			if( bnd_hi(i)==1 .AND. bnd_lo(i)==1 ) then         ! Check for n(i) = 1
				wt_lo(i) = 0.5;  bin_lo(i) = 1    
				wt_hi(i) = 0.5;  bin_hi(i) = 1
			else
				wt_lo(i) = 0.0;  wt_hi(i) = 0.0
				bin_lo(i) = bnd_hi(i) - 1
				bin_hi(i) = bnd_hi(i)
			end if
		end if
	end do

end subroutine get_interpn_weights
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
integer function get_nmod( k, nsurf) result(kmod)
	integer, intent(IN) :: k, nsurf
	integer             :: j

	kmod=1; j=k+1
	do while( j > 1)
		kmod = kmod + nsurf*j
		j=j-1
	end do
end function get_nmod
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
! The model cube has weird dimensions.  Grab the slice location without a
! do loop.  Formula is not general.  should figure it out eventually.
! Not implemented for nsurf > 2.
! ---------------------------------------------------------------------------
integer function get_cubeloc( order, nsurf, surf, split) result(kcube)
	integer, intent(IN) :: order, nsurf, surf, split
	if(order==0) then
		kcube = 1
	else
		kcube = nsurf*(order-1)*(order+2)/2 + (surf-1)*(order+1) + split + 2
	end if
end function get_cubeloc
! ---------------------------------------------------------------------------

! ---------------------------------------------------------------------------
! Inverse operation of get_cubeloc()
! ---------------------------------------------------------------------------
subroutine get_vectloc( kcube, maxorder, nsurf, order, surf, split)
	integer, intent(IN)  :: kcube, maxorder, nsurf
	integer, intent(OUT) :: order, surf, split
	integer :: ct, iorder, isurf, isplit, nnsurf

	ct = 0
	do iorder=0,maxorder
		if( iorder==0) then
			nnsurf=1
		else
			nnsurf=nsurf
		end if
		do isurf=1,nnsurf
			do isplit=0,iorder
				ct = ct + 1
				if( ct == kcube ) then
					order = iorder
					surf = isurf
					split = isplit
				end if	
			end do
		end do
	end do

end subroutine get_vectloc
! ---------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! Given a CMP gather, return 2-D logical array, T for missing traces, F otherwise
!--------------------------------------------------------------------------------
subroutine get_misstrace( nt, nx1, nx2, data, misstrace)
	integer,                 intent(IN)  :: nt, nx1, nx2
	real,    dimension(:),   intent(IN)  :: data
	logical, dimension(:,:), intent(OUT) :: misstrace

	integer :: ix1, ix2, j0, j1

	misstrace = .false.
	do ix2=1,nx2
		do ix1=1,nx1
			j0 = grid2line3( (/nt,nx1,nx2/), (/1 ,ix1,ix2/) )	
			j1 = grid2line3( (/nt,nx1,nx2/), (/nt,ix1,ix2/) )	
			if( sum(data(j0:j1)) == 0.0 ) misstrace(ix1,ix2) = .true.
		end do
	end do

end subroutine get_misstrace
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
! ADD NSURF
!--------------------------------------------------------------------------------
subroutine model_stats( label, model, nmod, maxorder, nsurf, nt, nx1, nx2,      &
                        ithr, jm0                                               )

	real,    dimension(:), intent(IN) :: model
	integer,               intent(IN) :: nmod, maxorder, nsurf, nt, nx1, nx2
	integer                           :: iord, isplit, k, i0, i1, ithr, jm0, jm1, &
	                                     nnsurf, isurf
	character(len=*),      intent(IN) :: label
	optional                          :: ithr, jm0
	logical                           :: blkwrite

	if(.NOT. present(ithr)) ithr=1
	blkwrite = .FALSE.
	if(present(jm0)) then
		blkwrite=.TRUE.
		jm1 = jm0 + nmod*nt*nx1*nx2
	end if

	do iord=0,maxorder
		if(iord==0) then
			nnsurf = 1
		else
			nnsurf = nsurf
		end if
		do isurf=1,nnsurf
			do isplit=0,iord

				k = get_cubeloc( iord, nsurf, isurf, isplit)

				i0 = grid2line4( (/nt,nx1,nx2,nmod/), (/1, 1,  1,  k/) )
				i1 = grid2line4( (/nt,nx1,nx2,nmod/), (/nt,nx1,nx2,k/) )
				if( blkwrite) then
					write(0,"(a,i2,a,a,a,i1,a,i1,a,i1,a)") 'Thr(',ithr,') :',trim(label), &
					                      ': ord/spl/srf(',iord,',',isplit,',',isurf,')...' 
					write(0,"(a,i7,a,i7,a,f,a,f)") '   ...min/max[modelblock(',jm0+i0,    &
					       ':',jm1+i1,')]: ', minval(model(i0:i1)),',',maxval(model(i0:i1))
				else
					write(0,"(a,i2,a,a,a,i1,a,i1,a,i1,a,i7,a,i7,a,f,a,f)")                &
					        'Thr(',ithr,') :',trim(label),': ord/spl(',iord,',',isplit,   &
					        '): min/max[model(',i0,':',i1,')]:',minval(model(i0:i1)),',', &
					        maxval(model(i0:i1))
				end if

			end do
		end do
	end do

end subroutine model_stats
!--------------------------------------------------------------------------------

!--------------------------------------------------------------------------------
subroutine data_stats( label, data, ithr, jd0, jd1)
	real,    dimension(:), intent(IN) :: data
	integer                           :: ithr, jd0, jd1
	character(len=*),      intent(IN) :: label
	optional                          :: ithr, jd0, jd1
	logical                           :: blkwrite

	if(.NOT. present(ithr)) ithr=0

	blkwrite = .FALSE.
	if( present(jd0) .AND. present(jd1) ) blkwrite=.TRUE.

	if(blkwrite) then
		write(0,"(a,i2,a,a,a,i,a,i,a,f,a,f)") 'Thr(',ithr,') :',trim(label),&
		   ': min/max[datablock(',jd0,':',jd1,')]: ', minval(data),',',maxval(data)
	else
		write(0,"(a,i2,a,a,f,a,f)") 'Thr(',ithr,') :',trim(label), &
		   ': min/max[data]: ', minval(data),',',maxval(data)
	end if

end subroutine data_stats
!--------------------------------------------------------------------------------

end module mult_nmo_util
