!-------------------------------------------------------------------------------
!  radial_trace.f90 - Created by Morgan Brown
!           Mon Oct 13 11:32:22 2003.
!-------------------------------------------------------------------------------
!       Usage: radial_trace.mod 
!  Parameters: 
! Description: 
!-------------------------------------------------------------------------------
module radial_trace

use sep
use adj_mod
use dottest
use morgan_util
use mult_nmo_util
use weight
use chain0_mod

implicit none

integer, private, save :: nt, nx1, nx2, nv1, nv2
real,    private, save :: dt, dx1, dx2, dv1, dv2, t0, x01, x02, v01, v02 
logical, private, save :: verbose, vinterp, lint, norm
logical, parameter, private, save :: T=.true., F=.false.
real, dimension(:), pointer, private, save :: wt, tmp

contains

!-------------------------------------------------------------------------------
subroutine radial_trace_init( nt_in, nx1_in, nx2_in, nv1_in, nv2_in, dt_in,    &
                              dx1_in, dx2_in, dv1_in, dv2_in, t0_in, x01_in,   &
                              x02_in, v01_in, v02_in, vinterp_in, lint_in,     &
                              norm_in, verb                                    )

	integer  :: nt_in, nx1_in, nx2_in, nv1_in, nv2_in, stat
	real     :: dt_in, dx1_in, dx2_in, dv1_in, dv2_in, t0_in, x01_in, x02_in,    &
	            v01_in, v02_in
	logical  :: vinterp_in, lint_in, norm_in, verb
	optional :: verb

	nt = nt_in;  nx1 = nx1_in;  nx2 = nx2_in;  nv1 = nv1_in;  nv2 = nv2_in
	dt = dt_in;  dx1 = dx1_in;  dx2 = dx2_in;  dv1 = dv1_in;  dv2 = dv2_in
	t0 = t0_in;  x01 = x01_in;  x02 = x02_in;  v01 = v01_in;  v02 = v02_in

	vinterp = vinterp_in
	lint = lint_in
	norm = norm_in

	verbose=F;  if(present(verb)) verbose=verb

	if( vinterp .AND. norm) then
		allocate( wt(nt*nv1*nv2), tmp(nt*nx1*nx2) )
		tmp = 1.0
		stat = radial_trace_vinterp_exec( T, F, wt, tmp)
		where( wt/=0.0 ) wt = 1.0/wt
		write(0,*) 'min/max/#nonzero(wt)',minval(wt),maxval(wt),count(wt/=0.0)
		deallocate( tmp)
		allocate( tmp(nt*nv1*nv2) )
	else if( (.NOT. vinterp) .AND. norm) then
		allocate( wt(nt*nx1*nx2), tmp(nt*nv1*nv2) )
		tmp = 1.0
		stat = radial_trace_xinterp_exec( F, F, tmp, wt)
		where( wt/=0.0 ) wt = 1.0/wt
		write(0,*) 'min/max/#nonzero(wt)',minval(wt),maxval(wt),count(wt/=0.0)
		deallocate( tmp)
		allocate( tmp(nt*nx1*nx2) )
	end if

	
end subroutine radial_trace_init
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
subroutine radial_trace_close( )
	if(norm) deallocate( wt, tmp)
end subroutine radial_trace_close
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! If applying just forward operator or adjoint, should normalize.  
! x-interpolation and v-interpolation should be normalized in different ways.
!
! x-interpolation forward operator:  W*R*m = d  where W = diag(w) and w = R*1
! v-interpolation forward operator:  R*W*m = d  where W = diag(w) and w = R'*1
!
!-------------------------------------------------------------------------------
integer function radial_trace_exec( adj, add, model, data)
	logical            :: adj, add 
	real, dimension(:) :: model, data 
	integer :: stat

	call adjnull( adj, add, model, data)

	if( verbose) call oper_stats( 'radial_trace_exec', T, model, data, adj)

	if( vinterp) then
		if( norm .AND. (.NOT. adj) ) then
			call weight_init( wt)
			call chain0( radial_trace_vinterp_exec, weight_lop, adj, add, model,     &
			             data, tmp                                                   )
		else
			stat = radial_trace_vinterp_exec( adj, add, model, data)	
		end if
	else
		if( norm .AND. (.NOT. adj) ) then
			call weight_init( wt)
			call chain0( weight_lop, radial_trace_xinterp_exec, adj, add, model,     &
			             data, tmp                                                   )
		else
			stat = radial_trace_xinterp_exec( adj, add, model, data)	
		end if
	end if

	if( verbose) call oper_stats( 'radial_trace_exec', F, model, data, adj)

	radial_trace_exec = 0

end function radial_trace_exec
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
integer function radial_trace_vinterp_exec( adj, add, model, data)
	logical            :: adj, add 
	real, dimension(:) :: model, data 
	real    :: x2, x1, tt, v1, v2, wlo(2), whi(2), xa(2), da(2), oa(2), w(4)
	integer :: ix2, ix1, it, ilo(2), ihi(2), na(2), im(4), nda(3), ida(3),        &
	           nva(2), iva(2), i, i0, k, za(2)

	call adjnull( adj, add, model, data)

	if( verbose) call oper_stats( 'radial_trace_vinterp_exec', T, model, data, adj)

	za = (/1,1/)
	na = (/nv1,nv2/)
	da = (/dv1,dv2/)
	oa = (/v01,v02/)
	nda = (/nt,nx1,nx2/)
	nva = (/nt,nv1*nv2/)

	do ix2=1,nx2
		x2 = (ix2-1)*dx2 + x02
		do ix1=1,nx1
			x1 = (ix1-1)*dx1 + x01
			do it=1,nt
				tt = (it-1)*dt + t0
				if( tt==0.0 ) cycle
				v1 = x1/tt
				v2 = x2/tt
				xa = (/v1,v2/)
				ida = (/it,ix1,ix2/)

				if( nx2>1) then
					call get_interpn_weights( xa, da, oa, ilo, ihi, wlo, whi, za, na)
				else
					call get_interp1_weights( v1, dv1, v01, ilo(1), ihi(1), wlo(1), whi(1), 1, nv1)
					ilo(2) = 1;   ihi(2) = 1
					wlo(2) = 0.5; whi(2) = 0.5
				end if
				call bins_weights_to_1d( na, wlo, whi, ilo, ihi, w, im)
				i0 = grid2line3( nda, ida)

				do i=1,4
					iva = (/it,im(i)/)
					k = grid2line2( nva, iva)               ! averaging across the k (model)
					if( adj) then
						model(k) = model(k) + w(i)*data(i0)
					else
						data(i0) = data(i0) + w(i)*model(k)
					end if
				end do

			end do
		end do
	end do

	if( verbose) call oper_stats( 'radial_trace_vinterp_exec', F, model, data, adj)

	radial_trace_vinterp_exec = 0

end function radial_trace_vinterp_exec
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
integer function radial_trace_xinterp_exec( adj, add, model, data)
	logical            :: adj, add 
	real, dimension(:) :: model, data 
	real    :: x2, x1, tt, v1, v2, wlo(2), whi(2), xa(2), da(2), oa(2), w(4)
	integer :: iv2, iv1, it, ilo(2), ihi(2), na(2), im(4), nda(3), ida(3),        &
	           nva(2), iva(2), i, i0, k, za(2)

	call adjnull( adj, add, model, data)

	if( verbose) call oper_stats( 'radial_trace_xinterp_exec', T, model, data, adj)

	za = (/1,1/)
	na = (/nx1,nx2/)
	da = (/dx1,dx2/)
	oa = (/x01,x02/)
	nda = (/nt,nv1,nv2/)
	nva = (/nt,nx1*nx2/)

	do iv2=1,nv2
		v2 = (iv2-1)*dv2 + v02
		do iv1=1,nv1
			v1 = (iv1-1)*dv1 + v01
			do it=1,nt
				tt = (it-1)*dt + t0
				x1 = v1*tt
				x2 = v2*tt
				xa = (/x1,x2/)
				ida = (/it,iv1,iv2/)

				if( nx2>1) then
					call get_interpn_weights( xa, da, oa, ilo, ihi, wlo, whi, za, na)
				else
					call get_interp1_weights( x1, dx1, x01, ilo(1), ihi(1), wlo(1), whi(1), 1, nx1)
					ilo(2) = 1;   ihi(2) = 1
					wlo(2) = 0.5; whi(2) = 0.5
				end if
				call bins_weights_to_1d( na, wlo, whi, ilo, ihi, w, im)
				i0 = grid2line3( nda, ida)

				do i=1,4
					iva = (/it,im(i)/)
					k = grid2line2( nva, iva)               ! averaging across the k (data)
					if( adj) then
						model(i0) = model(i0) + w(i)*data(k)
					else
						data(k) = data(k) + w(i)*model(i0)
					end if
				end do

			end do
		end do
	end do

	if( verbose) call oper_stats( 'radial_trace_xinterp_exec', F, model, data, adj)

	radial_trace_xinterp_exec = 0

end function radial_trace_xinterp_exec
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
!subroutine radial_trace_test( )
!	real    :: dot1(2), dot2(2)
!	integer :: nmod, ndat
!
!	call radial_trace_init( )
!	call dot_test( radial_trace_exec, nmod, ndat, dot1, dot2)
!end subroutine radial_trace_test
!-------------------------------------------------------------------------------

end module radial_trace
